-- Copyright (c) 2023-2024 Chris Pressey, Cat's Eye Technologies
-- This file is distributed under a BSD license.  See LICENSES directory:
-- SPDX-License-Identifier: LicenseRef-BSD-2-Clause-X-Fountain

module Language.Fountain.Generator (constructState, generateFrom, obtainResult) where

import Data.Maybe (mapMaybe)
import System.Random (StdGen, mkStdGen)

import Language.Fountain.Randomness (shuffle)
import Language.Fountain.Value
import Language.Fountain.Constraint
import Language.Fountain.Store (Store, empty, updateStore, trace)
import Language.Fountain.Unify (applyConstraint)
import Language.Fountain.Grammar
import Language.Fountain.Loader (constructStore)

--
-- In the code, we often call this variable `state`, or, if that
-- would shadow a variable already called `state`, we call it `st`
--
data GenState = Generating {
                  text :: String,
                  store :: Store Value,
                  prng :: StdGen
                }
    deriving (Show, Eq)

type GenResult = Either String GenState

--
-- Utils
--

can (Just _) = True
can Nothing  = False

genTerminal :: Char -> GenState -> GenState
genTerminal c state@Generating{ text=text } = state{ text=(c:text) }

--
-- Alt choices need preconditions during generation because
-- we need some guidance of which one to pick.
--
getPreCondition :: Expr -> Maybe (Constraint Value)
getPreCondition (Seq (x:_)) = getPreCondition x
getPreCondition (Constraint c) = Just c
getPreCondition _ = Nothing

missingPreConditions choices =
    mapMaybe (\x -> case getPreCondition x of
        Just _ -> Nothing
        Nothing -> Just x
      ) choices

scramble :: [a] -> GenState -> ([a], GenState)
scramble choices state@Generating{ prng=prng } =
    let
        (choices', prng') = shuffle choices $ prng
    in
        (choices', state{ prng=prng' })

--
-- Generator
--

gen :: Grammar -> GenState -> Expr -> GenResult

gen g state (Seq s) = genSeq state s where
    genSeq st [] = Right st
    genSeq st (e : rest) =
        case gen g st e of
            Left x    -> Left x
            Right st' -> genSeq st' rest

-- Hello, Mrs Backtracking Alternation!
gen g state (Alt (Just mode) choices) =
    -- Note, we try all the possibilities here, regardless of their preconditions.
    -- We also scramble them, unless determinism during backtracking was requested.
    let
        (choices', state') = case mode of
            Deterministic    -> (choices, state)
            NonDeterministic -> scramble choices state
    in
        genAlt state' choices' where
            genAlt _st [] = Left "no more choices"
            genAlt st (e : rest) =
                case gen g st e of
                    Right st' -> Right st'
                    Left _    -> genAlt st rest

-- Hello, Mrs Non-Backtracking Alternation!
gen g state@Generating{ store=store } (Alt Nothing choices) =
    -- We look at all the choices; each should start with a pre-condition
    -- determining whether we can select it; and we should narrow down our
    -- choices based on that.
    case missingPreConditions choices of
        missing@(_:_) ->
            error ("No pre-condition present on these Alt choices: " ++ (depictExprs missing))
        [] ->
            let
                preConditionedChoices = map (\x -> (getPreCondition x, x)) choices
                isApplicableChoice (Just c, _) = can $ applyConstraint c store
                isApplicableChoice _ = False
                applicableChoices = filter (isApplicableChoice) preConditionedChoices
            in
                genAlt state applicableChoices
            where
                genAlt _st [] = Left "no more choices"
                -- we ignore the constraint here because it will be found and applied when we descend into e
                genAlt st [(_, e)] = gen g st e
                genAlt _st other =
                    error ("Multiple pre-conditions are satisfied in Alt: " ++ (depictExprs (map (snd) other)))


gen _g _state (Loop _ []) = error "No postconditions defined for this Loop"
gen g state (Loop l postconditions) = genLoop state l where
    genLoop st e =
        case gen g st e of
            -- If the loop body fails, we should fail too.
            Left x -> Left x
            Right st'@Generating{ store=store } ->
                case checkLimit postconditions store of
                    -- All postconditions met, terminate the loop.  (Note that we throw away
                    -- the result of applying the constraints here, because they will follow
                    -- this loop directly, and we don't want to apply them twice.)
                    Just _  -> Right st'{ store=trace ("Done " ++ (show postconditions)) store }
                    -- Not all postconditions met -- go 'round again
                    Nothing -> genLoop st' e
    checkLimit [] st = Just st
    checkLimit (c:cs) st =
        case applyConstraint c st of
            Nothing -> Nothing
            Just st' -> checkLimit cs st'

gen _g st (Terminal c) = Right $ genTerminal c st
gen g st@Generating{ store=store } (NonTerminal nt actuals) =
    let
        formals = getFormals nt g
        newStore = updateStore actuals formals store empty
        st' = st{ store=newStore }
        expr' = production nt g
    in
        case gen g st' expr' of
            Right st''@Generating{ store=modifiedStore } ->
                -- After calling the NT, reconcile the store with the returned modifications. 
                Right $ st''{ store=updateStore formals actuals modifiedStore store }
            Left x ->
                Left x

gen _g st@Generating{ store=store } (Constraint cstr) =
    case applyConstraint cstr store of
        Just store' ->
            Right $ st{ store=trace ("OK " ++ depictConstraint cstr) store' }
        Nothing ->
            Left "constraint failed"

--
-- Usage interface
--

constructState :: Int -> [String] -> GenState
constructState seed initialParams =
    Generating "" (constructStore initialParams) (mkStdGen seed)

generateFrom :: Grammar -> String -> GenState ->  GenResult
generateFrom g start state = gen g state (production start g)

obtainResult :: GenResult -> Either String String
obtainResult (Right (Generating s _ _)) = Right $ reverse s
obtainResult (Left x) = Left x
