-- Copyright (c) 2023-2024 Chris Pressey, Cat's Eye Technologies
-- This file is distributed under a BSD license.  See LICENSES directory:
-- SPDX-License-Identifier: LicenseRef-BSD-2-Clause-X-Fountain

{-# LANGUAGE FlexibleContexts #-}
module Language.Fountain.Loader where

--
-- In a more conventional project, this module would be called "Parser",
-- but that would be somewhat too confusing here.
--

import Data.Char (chr)
import Text.ParserCombinators.Parsec

import Language.Fountain.Value
import Language.Fountain.Store
import Language.Fountain.Grammar
import Language.Fountain.Constraint


fountain = do
    fspaces
    ps <- many prod
    fspaces
    eof
    return (Grammar ps)

prod = do
    nt <- capWord
    p <- option [] formals
    mode <- prodMode
    keyword "::="
    e <- prodExpr0 mode
    keyword ";"
    return Production{ ntname=nt, params=p, backtrackMode=mode, constituents=e }

prodMode = try det <|> try nondet <|> (return Nothing)

det = do
    keyword "("
    keyword "*"
    keyword ")"
    keyword "("
    keyword "!"
    keyword ")"
    return $ Just Deterministic

nondet = do
    keyword "("
    keyword "*"
    keyword ")"
    return $ Just NonDeterministic

formals = do
    keyword "<"
    v <- sepBy (variable) (keyword ",")
    keyword ">"
    return v

prodExpr0 bt = do
    es <- sepBy (prodExpr1 bt) (keyword "|")
    return $ Alt bt es

prodExpr1 bt = do
    es <- many1 $ term bt
    return $ Seq $ flattenseq es where
        flattenseq [] = []
        flattenseq (s:ss) = case s of
            -- Note that xs will always be flat already
            Seq xs -> xs ++ (flattenseq ss)
            _      -> (s:flattenseq ss)

term bt = (try $ parenExpr bt) <|> (try $ loopExpr bt) <|> (try $ constraintExpr bt) <|> (try $ terminal bt) <|> nonterminal bt

parenExpr bt = do
    keyword "("
    e <- prodExpr0 bt
    keyword ")"
    return e

loopExpr bt = do
    keyword "{"
    e <- prodExpr0 bt
    keyword "}"
    return $ Loop e []

constraintExpr _bt = do
    keyword "<."
    c <- constrainer
    keyword ".>"
    return $ Constraint $ c

terminal _bt = do
    s <- quotedString <|> charlit
    case s of
        [c] -> return $ Terminal $ c
        _ -> return $ Seq $ map (\c -> Terminal c) s

nonterminal _bt = do
    s <- capWord
    a <- option [] actuals
    return $ NonTerminal s a
    where
        actuals = do
            keyword "<"
            v <- sepBy (variable) (keyword ",")
            keyword ">"
            return v

constrainer = (try unify) <|> (try inc) <|> (try dec) <|>
                (try gte) <|> (try gt) <|> (try lte) <|> (try lt) <|> (try both)

unify = do
    v <- variable
    keyword "="
    n <- conExpr
    return $ Unify v n

inc = do
    v <- variable
    keyword "+="
    e <- conExpr
    return $ Inc v e

dec = do
    v <- variable
    keyword "-="
    e <- conExpr
    return $ Dec v e

gte = do
    v <- variable
    keyword ">="
    e <- conExpr
    return $ GreaterThanOrEqual v e

gt = do
    v <- variable
    keyword ">"
    e <- conExpr
    return $ GreaterThan v e

lte = do
    v <- variable
    keyword "<="
    e <- conExpr
    return $ LessThanOrEqual v e

lt = do
    v <- variable
    keyword "<"
    e <- conExpr
    return $ LessThan v e

both = do
    -- TODO: this syntax is awful, change it please
    keyword "&&"
    c1 <- constrainer
    keyword ","
    c2 <- constrainer
    return $ Both c1 c2

variable = do
    s <- lowWord
    return $ Var s

conExpr = (try conLiteral) <|> varExpr

conLiteral = (try intExpr) <|> stringExpr

intExpr = do
    i <- intlit
    return $ ConVal (IntVal i)

stringExpr = do
    s <- quotedString
    return $ ConVal (StrVal s)

varExpr = do
    v <- variable
    return $ ConVar v

--
-- Low level: Concrete things
--

keyword s = do
    _ <- try (string s)
    fspaces

capWord = do
    c <- upper
    s <- many (alphaNum)
    fspaces
    return (c:s)

lowWord = do
    c <- lower
    s <- many (alphaNum)
    fspaces
    return (c:s)

intlit = do
    sign <- option 1 leadingMinus
    c <- digit
    cs <- many digit
    fspaces
    return ((read (c:cs) :: Integer) * sign)

leadingMinus :: Parser Integer
leadingMinus = do
    _ <- char '-'
    return (-1)

quotedString = do
    _ <- char '"'
    s <- many $ satisfy (\x -> x /= '"')
    _ <- char '"'
    fspaces
    return s

charlit = do
    _ <- char '#'
    c <- digit
    cs <- many digit
    fspaces
    return [chr (read (c:cs) :: Int)]

fspaces = do
    spaces
    _ <- many comment
    return ()

comment = do
    _ <- string "//"
    _ <- many $ satisfy (\x -> x /= '\n')
    (do { _ <- char '\n'; return ()} <|> eof)
    fspaces
    return ()

--
-- Driver
--

parseFountain :: String -> Either ParseError Grammar
parseFountain text = parse fountain "" text

parseConstConstraint :: String -> (Variable, Value)
parseConstConstraint text = case parse unify "" text of
    Right (Unify v (ConVal i)) -> (v, i)
    v -> error ("parseConstConstraint: " ++ show v)

constructStore :: [String] -> Store Value
constructStore [] = empty
constructStore (constConstrainer:rest) =
    let
        (k, v) = parseConstConstraint constConstrainer
    in
        insert k v $ constructStore rest
