-- Copyright (c) 2023-2024 Chris Pressey, Cat's Eye Technologies
-- This file is distributed under a BSD license.  See LICENSES directory:
-- SPDX-License-Identifier: LicenseRef-BSD-2-Clause-X-Fountain

module Language.Fountain.Parser (constructState, parseFrom, obtainResult) where

import Data.Maybe (mapMaybe)

import Language.Fountain.Value
import Language.Fountain.Constraint
import Language.Fountain.Store (Store, empty, updateStore)
import Language.Fountain.Unify (applyConstraint)
import Language.Fountain.Grammar
import Language.Fountain.Loader (constructStore)


data ParseState = Parsing String (Store Value)
    deriving (Show, Ord, Eq)

type ParseResult = Either String ParseState

--
-- Utils
--

expectTerminal :: Char -> ParseState -> ParseResult
expectTerminal tc (Parsing (c:cs) a)
    | c == tc   = Right $ Parsing cs a
    | otherwise = Left "no match"
expectTerminal _tc (Parsing [] _a) =
    Left "end of string"

--
-- Alt choices need preconditions during parsing because it helps
-- efficiency by preventing unnecessary backtracking.  But note that
-- we need a more refined notion in parsing than in generation,
-- because a terminal counts as a precondition.
--
getPreCondition :: Expr -> Maybe (Constraint Value)
getPreCondition (Seq (x:_)) = getPreCondition x
getPreCondition (Constraint c) = Just c
getPreCondition (Terminal c) = Just $ Lookahead s where s = [c]
getPreCondition _ = Nothing

missingPreConditions choices =
    mapMaybe (\x -> case getPreCondition x of
        Just _ -> Nothing
        Nothing -> Just x
      ) choices

getApplicableChoices state choices =
    let
        preConditionedChoices = map (\x -> (getPreCondition x, x)) choices
        isApplicableChoice (Just c, _) = case applyParseConstraint c state of
            Left _ -> False
            Right _ -> True
        isApplicableChoice _ = False
    in
        filter (isApplicableChoice) preConditionedChoices

--
-- Parser
--

parse :: Grammar -> ParseState -> Expr -> ParseResult

parse g state (Seq s) = parseSeq state s where
    parseSeq st [] = Right st
    parseSeq st (e : rest) =
        case parse g st e of
            Right st' -> parseSeq st' rest
            Left x    -> Left x

-- Hello, Mrs Backtracking Alternation!
parse g state (Alt (Just _) choices) = parseAlt state choices where
    -- Note, we try all the possibilities here, regardless of their preconditions.
    -- We also don't care if the backtracking is deterministic or nondeterministic.
    parseAlt _st [] = Left "no more choices"
    parseAlt st (e : rest) =
        case parse g st e of
            Left _    -> parseAlt st rest
            Right st' -> Right st'

-- Hello, Mrs Non-Backtracking Alternation!
parse g state (Alt Nothing choices) =
    case missingPreConditions choices of
        missing@(_:_) ->
            error ("No pre-condition present on these Alt choices: " ++ (depictExprs missing))
        [] ->
            parseAlt state (getApplicableChoices state choices)
            where
                parseAlt _st [] = Left "no more choices"
                -- we ignore the constraint here because it will be found and applied when we descend into e
                parseAlt st [(_, e)] = parse g st e
                parseAlt (Parsing _str store) other =
                    error ("Multiple pre-conditions are satisfied in Alt: " ++ (depictExprs (map (snd) other)) ++ ", with state: " ++ show store)

parse g state (Loop l []) = parseLoop state l where
    parseLoop st e =
        case parse g st e of
            Right st' -> parseLoop st' e
            Left _    -> Right st

parse _g _state (Loop l (_:_)) = error ("Parsing can't handle decorated Loops: " ++ depictExpr l)

parse _g state (Terminal c) = expectTerminal c state

parse g (Parsing text store) (NonTerminal nt actuals) =
    let
        formals = getFormals nt g
        newStore = updateStore actuals formals store empty
        st' = Parsing text newStore
        expr' = production nt g
    in
        case parse g st' expr' of
            Right (Parsing text' modifiedStore) ->
                let
                    reconciledStore = updateStore formals actuals modifiedStore store
                in
                    Right $ Parsing text' reconciledStore
            Left x ->
                Left x

parse _g state (Constraint cstr) = applyParseConstraint cstr state


applyParseConstraint :: Constraint Value -> ParseState -> ParseResult
applyParseConstraint (Lookahead s) state@(Parsing (c:_) _) =
    if s == [c] then (Right state) else (Left "lookahead failed")
applyParseConstraint (Lookahead _) _ = Left "lookahead failed"
applyParseConstraint other (Parsing s store) =
    case applyConstraint other store of
        Just store' ->
            Right $ Parsing s store'
        Nothing ->
            Left "constraint failed"

--
-- Usage interface
--

constructState :: String -> [String] -> ParseState
constructState text initialParams = Parsing text $ constructStore initialParams

parseFrom :: Grammar -> String -> ParseState -> ParseResult
parseFrom g start st = parse g st (production start g)

obtainResult :: ParseResult -> Either String String
obtainResult (Right (Parsing s _)) = Right s
obtainResult (Left other) = Left other
