-- Copyright (c) 2023-2024 Chris Pressey, Cat's Eye Technologies
-- This file is distributed under a BSD license.  See LICENSES directory:
-- SPDX-License-Identifier: LicenseRef-BSD-2-Clause-X-Fountain

module Language.Fountain.Unify (
    applyConstraint
) where

import Language.Fountain.Value
import Language.Fountain.Constraint
import Language.Fountain.Store


--
-- Obtain the value of a constraint expression, relative to a Store.
--
conEval :: ConExpr a -> Store a -> Maybe a
conEval (ConVal val) _ = Just val
conEval (ConVar v) st = fetch v st

--
-- Apply a constraint, relative to a store, possibly updating the store
-- (by unification, causing the constraint to become true.)
--
applyConstraint :: Constraint Value -> Store Value -> Maybe (Store Value)
applyConstraint (Unify v (ConVal x)) st =
    case fetch v st of
        Just value ->
            if value == x then Just st else Nothing
        Nothing ->
            Just $ insert v x st
applyConstraint (Unify v (ConVar w)) st =
    case (fetch v st, fetch w st) of
        (Just vValue, Just wValue) ->
            if vValue == wValue then Just st else Nothing
        (Just vValue, Nothing) ->
            Just $ insert w vValue st
        (Nothing, Just wValue) ->
            Just $ insert v wValue st
        (Nothing, Nothing) ->
            Just st
applyConstraint (Inc v e) st =
    case (fetch v st, conEval e st) of
        (Just (IntVal vValue), Just (IntVal delta)) ->
            Just $ insert v (IntVal (vValue + delta)) st
        _ ->
            Nothing
applyConstraint (Dec v e) st =
    case (fetch v st, conEval e st) of
        (Just (IntVal vValue), Just (IntVal delta)) ->
            Just $ insert v (IntVal (vValue - delta)) st
        _ ->
            Nothing
applyConstraint (Both c1 c2) st =
    case applyConstraint c1 st of
        Just st' ->
            applyConstraint c2 st'
        Nothing ->
            Nothing
applyConstraint (GreaterThan v e) st = applyRelConstraint (>) v e st
applyConstraint (GreaterThanOrEqual v e) st = applyRelConstraint (>=) v e st
applyConstraint (LessThan v e) st = applyRelConstraint (<) v e st
applyConstraint (LessThanOrEqual v e) st = applyRelConstraint (<=) v e st
applyConstraint other _state = error ("Can't handle this: " ++ show other)

applyRelConstraint :: (Value -> Value -> Bool) -> Variable -> ConExpr Value -> Store Value -> Maybe (Store Value)
applyRelConstraint op var e st =
    case (fetch var st, conEval e st) of
        (Just value, Just target) ->
            if value `op` target then Just st else Nothing
        _ ->
            Nothing
